import functools
import inspect
from typing import TYPE_CHECKING, AnyStr, List

import regex as re

from ruleau.constants import OverrideLevel
from ruleau.docs import clean_source, comments, description, doctests, parameters, title
from ruleau.exceptions import (
    RuleIdIllegalCharacterException,
    RuleRequiresIdException,
    RuleRequiresNameException,
)

if TYPE_CHECKING:
    from ruleau.types import Function


class Rule:
    def __init__(
        self,
        func: "Function",
        id: AnyStr,
        name: AnyStr,
        depends_on: List["Rule"],
        override_level: OverrideLevel,
        lazy_dependencies: bool,
    ):
        """
        :param func: User defined rule
        :param name: User defined human readable name of the rule
        :param depends_on: Rule dependencies
        :param override_level: Override level
        :param lazy_dependencies: Flag to switch loading of rule dependencies lazily
        """
        self.id = id
        self.name = name
        # Validate the rule, make sure the name is always set for a rule
        self.validate()
        # Set the user defined function
        self.func = func
        self.depends_on = depends_on
        self.override_level = override_level
        self.__name__ = func.__name__
        self.lazy_dependencies = lazy_dependencies

        # This preserves the original Docstring on the decorated function
        # which allows DocTest to detect the function
        functools.update_wrapper(self, func)

    def __str__(self) -> str:
        return self.__name__

    def __call__(self, *args, **kwargs) -> bool:
        return self.func(*args, **kwargs)

    def validate(self):
        """
        Validator to check if top level rule has a human readable name and
        and id
        :raises: TopLevelRuleRequiresNameException
        :raises: RuleRequiresIdException
        """
        if not self.name or not isinstance(self.name, str):
            raise RuleRequiresNameException()
        if not self.id or not isinstance(self.id, str):
            raise RuleRequiresIdException()

        # Validate the Rule ID
        if not re.match(r"^([a-zA-Z0-9-_.~]+)+$", self.id):
            raise RuleIdIllegalCharacterException()

    def _get_source(self) -> AnyStr:
        return clean_source(inspect.getsource(self.func))

    @property
    def description(self) -> AnyStr:
        return description(self.func.__doc__)

    def parse(self):
        return {
            "function_name": self.__name__,
            "id": self.id,
            "name": self.name,
            "title": title(self.__name__),
            "override_level_name": self.override_level.name,
            "override_level": self.override_level.value,
            "source": self._get_source(),
            "comments": comments(self._get_source()),
            "docstring": self.func.__doc__,
            "description": self.description,
            "parameters": parameters(self.func.__doc__),
            "dependencies": [dependent.parse() for dependent in self.depends_on],
            "doctests": doctests(self.func),
        }



from typing import AnyStr, Dict

from .rule import Rule


class Process:
    """Class holding rule process"""

    def __init__(
        self,
        process_id: str,
        name: str,
        description: str,
        flattened_rules: Dict[AnyStr, Rule],
        root_rule: Rule,
    ):
        self.id = process_id
        self.name = name
        self.description = description
        self.flattened_rules = flattened_rules
        self.root_rule = root_rule

    def parse(self):
        rules = []
        for rule in self.flattened_rules:
            parsed_rule = rule["rule"].parse()
            rules.append(
                {
                    "id": rule["id"],
                    "name": rule["name"],
                    "description": parsed_rule["description"],
                    "override_level": parsed_rule["override_level"],
                    "order": rule["order"],
                    "dependencies": rule["dependencies"],
                    "doctests": parsed_rule["doctests"],
                    "parameters": [
                        {"name": name, "value": value}
                        for name, value in parsed_rule["parameters"].items()
                    ],
                }
            )
        return {
            "id": self.id,
            "name": self.name,
            "description": self.description,
            "root_rule": self.root_rule.id,
            "rules": rules,
        }



from typing import AnyStr

from ruleau.decorators import rule
from ruleau.execute import ExecutionResult


def Any(rule_id: AnyStr, name: AnyStr, *args):
    """Aggregator to implement OR operation
    Returns truthy, if any one of the rule result is truthy
    """

    @rule(rule_id=rule_id, name=name, depends_on=args)
    def any_aggregator(context: ExecutionResult, _):
        return any(result.result for result in context.dependant_results)

    return any_aggregator


def All(rule_id: AnyStr, name: AnyStr, *args):
    """Aggregator to implement AND operation
    Returns truthy, if and all of the rule results are truthy
    """

    @rule(rule_id=rule_id, name=name, depends_on=args)
    def all_aggregator(context: ExecutionResult, _):
        return all(result.result for result in context.dependant_results)

    return all_aggregator



from ruleau import All, execute, rule
from ruleau.adapter import ApiAdapter


# create a rule
@rule(rule_id="rul_01", name="over_18")
def over_18(context, payload):
    """
    Check applicant over 18
    >>> age(None, {"age": 18})
    True
    >>> age(None, {"age": 17})
    False
    """
    return "age" in payload and payload["age"] >= 18


@rule(rule_id="rul_02", name="not_short", depends_on=[over_18])
def not_short(context, payload):
    """
    Check applicant isn't short
    """
    return (
        context.dependant_results.over_18.result
        and "height" in payload
        and payload["height"] == "tall"
    )


collection = All(
    "Age Checker",
    "rul_03",
    All("sub level", "rul_04", over_18, All("sub sub level", "rul_05", not_short)),
)

# create a payload (the answers to the rule's questions)
payload = {"age": 18, "height": "tall", "case_id": "testing-xx"}
api_adapter = ApiAdapter(base_url="http://localhost:8000/")
# send the results
result = execute(
    collection,
    payload,
    # api_adapter=api_adapter,
    case_id_jsonpath="$.case_id",
)
print(f"result: {result.result}")



import argparse
import importlib
import re
import sys
from argparse import Namespace
from doctest import DocTestFinder
from os import mkdir, path
from textwrap import dedent
from typing import TYPE_CHECKING, AnyStr, Optional

from jinja2 import Template

if TYPE_CHECKING:
    from ruleau.rule import Rule
    from ruleau.types import Function

PARAM_REGEX_PATTERN = ":(?P<name>[\*\s\w]+): (?P<description>.*?)\Z"  # noqa: W605


def trim(docstring):
    """trim function implementation from PEP-257
    https://www.python.org/dev/peps/pep-0257/#id18

    :param docstring: Docstring for a function
    """
    if not docstring:
        return ""
    # Convert tabs to spaces (following the normal Python rules)
    # and split into a list of lines:
    lines = docstring.expandtabs().splitlines()
    # Determine minimum indentation (first line doesn't count):
    indent = sys.maxsize
    for line in lines[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))
    # Remove indentation (first line is special):
    trimmed = [lines[0].strip()]
    if indent < sys.maxsize:  # pragma: no cover
        for line in lines[1:]:
            trimmed.append(line[indent:].rstrip())
    # Strip off trailing and leading blank lines:
    while trimmed and not trimmed[-1]:
        trimmed.pop()
    while trimmed and not trimmed[0]:
        trimmed.pop(0)

    if "\n" in docstring:
        trimmed.append("")
    # Return a single string:
    return "\n".join(trimmed)


def description(docstring: AnyStr) -> AnyStr:
    """Takes a rule docstring, and returns just the lines
    representing a description

    :param docstring: Docstring for a function
    """
    result = trim(docstring or "")
    lines = result.split("\n")
    end_of_description_idx = next(
        (
            idx
            for idx, line in enumerate(lines)
            if any(
                [
                    # Line is empty i.e. new line after description
                    line == "" or line.isspace(),
                    # Line is the start of a doc string parameter
                    re.match(PARAM_REGEX_PATTERN, line),
                    # Line is the start of a doc test
                    line.startswith(">>>"),
                ]
            )
        ),
        # If no end to the description then it ends with the string
        len(lines),
    )

    description = "\n".join(lines[:end_of_description_idx])
    # Remove final new line character if present
    return description.removesuffix("\n")


def parameters(docstring: AnyStr):
    """Parse list of parameters out of the docstring

    :param docstring: The docstring of the rule
    :returns: [{"name": ..., "description": ...}, ...]
    """
    PARAM_REGEX = re.compile(PARAM_REGEX_PATTERN)

    params = {}
    if docstring:
        lines = docstring.split("\n")
        for line in lines:
            match = PARAM_REGEX.findall(line.strip())
            if match:
                param_name, param_description = match[0]
                params.update({param_name: param_description})

    return params


def comments(source: AnyStr) -> AnyStr:
    """Returns single line comments from provided source

    :param source: clean source (output of clean_source)
    :returns: returns single line comments from source code
    """
    output_lines = [
        line for line in source.splitlines() if line.strip().startswith("#")
    ]
    return "\n".join(output_lines)


def title(name: AnyStr) -> AnyStr:
    """Convert the function name into a title

    :param name: Name of a function
    """
    return name.replace("_", " ").title()


def doctests(func: "Function") -> [dict]:
    """Extract the doctest from a rule
    :param func: Function wrapped in a Rule instance
    :return: List of doctest examples
    """
    finder = DocTestFinder(verbose=False, recurse=False)
    return [
        {"test": example.source, "result": example.want}
        for tests in finder.find(func)
        for example in tests.examples
    ]


def order_rules(rules: ["Rule"]) -> ["Rule"]:
    """Order rules so top level rules appear first"""
    is_root = {rule: True for rule in rules}

    # If any rule is dependant on a rule then it is not the root
    for rule in rules:
        for dependant in rules:
            if dependant != rule and dependant in rule.depends_on:
                is_root[dependant] = False

    roots = [rule for rule, is_root_rule in is_root.items() if is_root_rule]
    others = [rule for rule, is_root_rule in is_root.items() if not is_root_rule]
    return roots + others


def get_start_of_docstring(source_lines: AnyStr) -> Optional[int]:
    """Get the index of line on which the doc string starts.

    :param source_lines: A list of the lines of the source of the function
    without the header or decorators
    """
    # Loop to find the start of the doc string
    for idx, line in enumerate(source_lines):

        # Skip any empty lines
        if line == "" or line.isspace():  # pragma: no cover
            # This is covered in the tests but no picked up by coverage
            continue

        # Find the start of the doc string
        if line.strip().startswith('"""'):
            return idx

        # If we find any non empty line that is not starting with """ then
        # there is no doc string for this function
        else:
            return None


def get_end_of_docstring(source_lines: AnyStr, start_idx=0) -> int:
    """Get the index of line on which the doc string ends.

    :param source_lines: A list of the lines of the source of the function
    without the header or decorators
    :param start_idx: The line on which the docstring starts

    :raise: Exception if no end of docstring found
    """
    for line_number, line in enumerate(source_lines[start_idx:]):  # pragma: no cover
        # A docstring cannot be started and not ended

        count = line.count('"""')  # Number of occurances of """ on line

        # If we are on the starting line (same as doc_string_start_idx) we
        # need to find 2 """ for the doc string to be closed.
        if (line_number == 0 and count == 2) or (line_number != 0 and count >= 1):
            return start_idx + line_number

    raise ValueError("No end of docstring found")


def clean_source(source: AnyStr) -> AnyStr:
    """Clean the source code.

    Remove all the lines until we have passed the decorators, function header
    and docstring.

    :param source: The source code to clean.
    """
    # Remove common indent
    source = dedent(source)

    #  Get lines from source
    lines = source.split("\n")

    # Get line function header starts on. All function sources must have a header
    header_index = next(
        idx for idx, line in enumerate(lines) if line.startswith("def")
    )  # pragma: no cover

    # Remove anything above header from lines (decorators) and header it self
    lines = lines[header_index + 1 :]  # noqa: E203

    # Remove the docstring
    doc_string_start_idx = get_start_of_docstring(lines)

    # If we found a doc string
    if doc_string_start_idx is not None:
        doc_string_end_idx = get_end_of_docstring(lines, doc_string_start_idx)
        # Remove the doc string lines
        lines = lines[doc_string_end_idx + 1 :]  # noqa: E203

    clean_source = dedent("\n".join(lines))
    return clean_source


def generate_documentation(rules) -> AnyStr:
    """Returns a HTML string, documenting the passed in rule and its dependants

    :param rules: List of rules to generate document of
    """
    with open(
        path.join(path.dirname(path.realpath(__file__)), "html", "documentation.html")
    ) as f:
        doc_template = Template(f.read())

    rules = order_rules(rules)
    enriched_rules = [rule.parse() for rule in rules]
    colors = {
        "light": "75FF96",
        "primary": "12C170",
        "dark": "17764A",
    }
    return doc_template.render(rules=enriched_rules, colors=colors)


def render_doc_for_module(module_file) -> AnyStr:
    """Render the documentation from a rule module
    :param module_file:
    :return:
    """
    spec = importlib.util.spec_from_file_location(module_file, module_file)
    config = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(config)

    # Pick rules to document
    rules = [
        rule
        for rule in vars(config).values()
        if getattr(type(rule), "__name__", None) == "Rule"
    ]
    if not rules:
        raise Exception("No rules found in {}.".format(module_file))

    return generate_documentation(rules)


def generate_and_save_to_file(input_files, output_dir) -> int:
    """Generate the documentation and save to a file in provided output dir
    :param input_files: Rule modules provided by user via CLI
    :param output_dir: Output directory of generated HTML file
    :return:
    """
    generated_docs = []

    if len(input_files) == 0:
        raise Exception("No file(s) supplied to generate documentation for.")

    if not path.exists(output_dir):
        mkdir(output_dir)

    for target_file in input_files:
        if not path.exists(target_file):
            raise ValueError(f"{target_file} does not exist")

        documentation = render_doc_for_module(target_file)

        output_filename = path.join(
            output_dir, "{}.html".format(path.basename(target_file).split(".")[0])
        )

        with open(output_filename, "w") as f:
            f.write(documentation)
        generated_docs.append(output_filename)

    print(
        f"{len(generated_docs)} doc{'s' if len(generated_docs) > 1 else ''} generated."
    )
    print("\n".join(generated_docs))

    return 0


def get_arguments(args) -> Namespace:
    """Parses arguments for the ruleau-docs command
    :param args:
    :return:
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--output-dir",
        dest="output_dir",
        default="./",
        help="Directory for generated documents",
    )
    parser.add_argument("files", nargs="*")
    return parser.parse_args(args)


def main() -> int:  # pragma: no cover
    """Console script for deft document generation.
    USAGE:
    ```bash
    $ ruleau-docs [--output-dir=<argument>] filename ...
    ```
    """
    args = get_arguments(sys.argv[1:])
    return generate_and_save_to_file(args.files, args.output_dir)


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



from ruleau import All, OverrideLevel, rule


@rule(rule_id="rul_01", name="kyc_threshold", override_level=OverrideLevel.NO_OVERRIDE)
def kyc_risk_greater_than_threshold(_, payload):
    """
    Know Your Customer (KYC) score must be greater than the threshold,
    in this case greater than `LOW`

    :Override Guidance: kyc_threshold to mark this failure into review

    >>> kyc_risk_greater_than_threshold(None, {"data": {"kyc": "HIGH"}})
    False
    >>> kyc_risk_greater_than_threshold(None, {"data": {"kyc": "LOW"}})
    True
    """
    return payload["data"]["kyc"] == "LOW"


@rule(rule_id="rul_02", name="fico_score")
def fico_score_greater_than_threshold(_, payload):
    """
    FICO score must be greater than 630

    :Owner: Penny Farthing
    :Override Guidance: Feel free to override in almost any circumstance

    >>> fico_score_greater_than_threshold(None, {"data": {"fico_score": 400}})
    False
    >>> fico_score_greater_than_threshold(None, {"data": {"fico_score": 630}})
    False
    >>> fico_score_greater_than_threshold(None, {"data": {"fico_score": 650}})
    True
    """
    return payload["data"]["fico_score"] > 630


@rule(rule_id="rul_03", name="no_ccjs")
def has_no_ccjs(_, payload):
    """
    Make sure customer has no county court judgements
    >>> has_no_ccjs(None, {"data": {"ccjs": []}})
    True
    >>> has_no_ccjs(None, {"data": {"ccjs": ["Example CCJ"]}})
    False
    >>> has_no_ccjs(None, {"data": {"ccjs": [{"example": "CCJ Object"}]}})
    False
    """
    return len(payload["data"]["ccjs"]) == 0


will_lend = All(
    "rul_04",
    "Top level",
    kyc_risk_greater_than_threshold,
    fico_score_greater_than_threshold,
    has_no_ccjs,
)

from ruleau import ApiAdapter, execute

if __name__ == "__main__":
    result = execute(
        will_lend,
        {"data": {"fico_score": 150, "ccjs": [], "kyc": "low"}},
        api_adapter=ApiAdapter(base_url="http://127.0.0.1:8000"),
        case_id="132",
    )
    print("Result", result.result)



import logging
from copy import deepcopy
from json import dumps
from typing import Any, AnyStr, Dict, Iterable, Optional

from jsonpath_ng import parse

from ruleau.adapter import ApiAdapter
from ruleau.constants import OverrideLevel
from ruleau.exceptions import (
    CannotOverrideException,
    CaseIdRequiredException,
    DuplicateRuleIdException,
    DuplicateRuleNameException,
)
from ruleau.process import Process
from ruleau.rule import Rule
from ruleau.structures import RuleauDict

logger = logging.getLogger(__name__)


class DependantResults:
    def __init__(
        self,
        case_id: AnyStr,
        dependants: Iterable[Rule],
        payload: Dict[AnyStr, Any],
        api_adapter: Optional[ApiAdapter] = None,
        lazy: bool = False,
    ):
        self.case_id = case_id
        self.dependants = {dep.__name__: dep for dep in dependants}
        self.payload = deepcopy(payload)
        self.api_adapter = api_adapter
        self.results = {}
        if not lazy:
            for depend in dependants:
                self.run(depend.__name__)

    def run(self, name):
        """
        Run and store the result of a rule dependency
        :param name:
        :return:
        """

        if name not in self.dependants:
            raise AttributeError(
                f"Result for rule '{name}' not available, as it was not "
                f"declared as a dependency. "
                f"depends_on={dumps(list(self.dependants.keys()))}"
            )
        # If the result of rule execution is not set, run & cache it
        if name not in self.results:
            self.results[name] = execute_rule(
                self.case_id,
                self.dependants[name],
                self.payload,
                self.api_adapter,
            )
        # Return the rule execution result
        return self.results[name]

    def __getattr__(self, name):
        # Get the attribute otherwise, run the dependency
        return getattr(super(), name, self.run(name))

    def __iter__(self):
        # Iterate through the dependencies
        for dep in self.dependants:
            yield getattr(self, dep)


class ExecutionResult:
    def __init__(
        self,
        executed_rule: Rule,
        payload: RuleauDict,
        result,
        dependant_results: DependantResults,
        override: AnyStr = None,
        original_result: Optional[bool] = None,
    ):
        self.executed_rule = executed_rule
        self.payload = payload
        self.result = result
        self.override = override
        self.original_result = original_result
        self.dependant_results = dependant_results


def apply_override(
    case_id,
    executable_rule: Rule,
    execution_result: ExecutionResult,
    api_adapter: ApiAdapter,
):
    # Get overrides for the rule in a case
    override = api_adapter.fetch_override(case_id, executable_rule)

    # Apply override to the executed rule result, if any
    # Overrides should only be applied to allowed rule and if they're present
    if override:
        # Throw an exception if the backend is trying to override a NO_OVERRIDE rule
        if executable_rule.override_level == OverrideLevel.NO_OVERRIDE:
            raise CannotOverrideException(f"Cannot override {executable_rule.name}")
        else:
            # Override the rule result and set the overridden flag
            execution_result.override = override["id"]
            execution_result.original_result = execution_result.result
            execution_result.result = override["applied"]
    return execution_result


def execute_rule(
    case_id: AnyStr,
    executable_rule: Rule,
    payload: Dict[AnyStr, Any],
    api_adapter: Optional[ApiAdapter] = None,
):
    api_result = {}
    # Create the rule result so that the execution can store the result
    if api_adapter:
        api_result = api_adapter.create_result(case_id, executable_rule)
    # Prep the rule payload
    rule_payload = RuleauDict(payload)
    # Prep the dependent results
    depend_results = DependantResults(
        case_id,
        executable_rule.depends_on,
        payload,
        api_adapter,
        lazy=executable_rule.lazy_dependencies,
    )
    # Prepare execution result for context from all dependencies
    context = ExecutionResult(executable_rule, rule_payload, None, depend_results)
    # Prepare execution result for the rule to be executed
    execution_result = ExecutionResult(
        executable_rule,
        rule_payload,
        executable_rule(context, rule_payload),
        depend_results,
    )
    # Store the rule result
    if api_adapter:
        # Apply overrides on the rule result
        execution_result = apply_override(
            case_id, executable_rule, execution_result, api_adapter
        )

        api_adapter.update_result(
            case_id, executable_rule, api_result, execution_result
        )
    # Return the rule result
    return execution_result


def flatten_rules(rule: Rule, flattened_rules: list, order: int = 0):
    """Flatten the rule to find it's dependencies
    :param rule:
    :param flattened_rules: Output to append flattened results to
    :param order: Execution Order of the rule
    :return:
    """
    dependencies = {
        "id": rule.id,
        "rule": rule,
        "name": rule.name or rule.__name__,
        "dependencies": [],
    }
    # Check if there are any rule dependencies
    if len(rule.depends_on):
        # If yes, find the rule IDs
        for dependency in rule.depends_on:
            # Recursively flatten rules
            flatten_rules(dependency, flattened_rules, order + 1)
            # Append the rule as a dependency
            dependencies["dependencies"].append(dependency.id)
    # Set the order of the current rule
    dependencies["order"] = order
    # Append th
    flattened_rules.append(dependencies)
    return flattened_rules


def flatten_rule_objects(rule: Rule, flat_rules=None) -> [Rule]:
    if flat_rules is None:
        flat_rules = []
    flat_rules.append(rule)
    if len(rule.depends_on):
        for dependency in rule.depends_on:
            flatten_rule_objects(dependency, flat_rules)
    return flat_rules


def validate_no_duplicate_rule_names(rules: [Rule]) -> None:
    """Returns True if there are no duplicate Rule Names are used
    A name can only be re-used if the same rule is included multiple times
    """
    rules_dict = {}
    for rule in rules:
        if rule.name not in rules_dict:
            rules_dict[rule.name] = rule
        else:
            if rule != rules_dict[rule.name]:
                raise DuplicateRuleNameException()


def validate_no_duplicate_rule_ids(rules: [Rule]) -> None:
    """Returns True if there are no duplicate Rule IDs used
    An ID can only be re-used if the same rule is included multiple times
    """
    rules_dict = {}
    for rule in rules:
        if rule.id not in rules_dict:
            rules_dict[rule.id] = rule
        else:
            if rule != rules_dict[rule.id]:
                raise DuplicateRuleIdException()


def execute(
    root_rule: Rule,
    payload: Dict[AnyStr, Any],
    case_id_jsonpath: AnyStr = None,
    case_id: Optional[AnyStr] = None,
    api_adapter: Optional[ApiAdapter] = None,
) -> ExecutionResult:
    """
    Executes the provided rule, following dependencies and
    passing in results accordingly
    """

    # If neither case_id_jsonpath or case_id are present, raise exception
    if not case_id_jsonpath and not case_id:
        raise CaseIdRequiredException()

    # If case_id is not present in parameters, find it
    if not case_id:
        case_id_results = parse(case_id_jsonpath).find(payload)
        if not case_id_results:
            raise ValueError("Case ID not found in payload")
        case_id = str(case_id_results[0].value)

    # If there's no case ID, don't run
    if not case_id:
        raise ValueError("Case ID not found")

    # Validate unique rule name
    flattened_rules_as_objects = flatten_rule_objects(root_rule)
    validate_no_duplicate_rule_names(flattened_rules_as_objects)

    # Validate unique rule ids
    validate_no_duplicate_rule_ids(flattened_rules_as_objects)

    # If API adapter was was passed sync the case

    if api_adapter:
        rules = flatten_rules(root_rule, [], 0)
        rules = {
            rule["id"]: rule for rule in rules
        }.values()  # Hot fix: De-duplicate rules
        process = Process(
            root_rule.id, root_rule.name, root_rule.description, rules, root_rule
        )

        # Sync the process rules
        api_adapter.sync_process(process)

        # Sync the case
        api_adapter.sync_case(case_id, process.id, payload)
    # Trigger the rule execution, from the top level rule
    return execute_rule(case_id, root_rule, payload, api_adapter)
